@@ -1,7 +1,8 @@
-# Copyright 2013-4, Sean B. Palmer
-# Source: http://inamidst.com/saxo/
+# http://inamidst.com/saxo/
+# Created by Sean B. Palmer
 
 import atexit
+import collections
 import configparser
 import imp
 import importlib
@@ -21,11 +22,13 @@ if "." in __name__:
     from . import scheduler
     from . import sqlite
     from .saxo import path as saxo_path
+    from .saxo import version as saxo_version
 else:
     import common
     import scheduler
     import sqlite
     from saxo import path as saxo_path
+    from saxo import version as saxo_version
 
 lock = threading.Lock()
 
@@ -37,6 +40,12 @@ def debug(*args, **kargs):
         except BrokenPipeError:
             sys.exit()
 
+def exit(code):
+    # TODO: Sock removal code
+    try: sys.exit(code)
+    finally: os._exit(code)
+    # TODO: Sometimes sys.exit doesn't work, not sure why
+
 # List of threads:
 # client.receive
 # client.send
@@ -124,6 +133,21 @@ class Message(object):
                 self.cmd = cmd
                 self.arg = arg
 
+    def authorised(self):
+        import re
+
+        config_owner = self.config.get("owner", "")
+        test_identity = False
+        if not "!" in config_owner:
+            test_identity = True
+            config_owner = config_owner + "!*@*"
+
+        mask = lambda g: "^" + re.escape(g).replace("\\*", ".*") + "$"
+        matches = re.match(mask(config_owner), self.prefix) is not None
+        if test_identity:
+            return matches and self.identified
+        return matches
+
     def client(self, *args):
         incoming.put(args)
 
@@ -139,12 +163,29 @@ def socket_receive(sock):
     try: receive_loop(sock, incoming)
     except Exception as err:
         # Usually IOError, EOFError, socket.error, or ssl.SSLError
-        ...
+        debug(str(err))
     incoming.put(("disco_receiving",))
 
+def flood_protection(recent):
+    recent.append(time.monotonic())
+    if len(recent) > 3:
+        recent.popleft()
+        if (recent[2] - recent[0]) < 2:
+            # debug("sleeping 1")
+            time.sleep(1)
+        elif (recent[2] - recent[1]) < 1:
+            # debug("sleeping 0.5")
+            time.sleep(0.5)
+        # else:
+        #     debug("sleeping 0")
+    else:
+        # debug("sleeping 0.25")
+        time.sleep(0.25)
+
 # threaded
 def socket_send(sock, flood=False):
     def sending(sock, flood=False):
+        recent = collections.deque()
         with sock.makefile("wb") as s:
             incoming.put(("sending",))
             while True:
@@ -156,9 +197,10 @@ def socket_send(sock, flood=False):
                 debug("->", repr(octets.decode("utf-8", "replace")))
                 s.write(octets)
                 s.flush()
+
                 if not flood:
-                    # TODO: Allow two or three burst lines
-                    time.sleep(1)
+                    flood_protection(recent)
+        # TODO: Surely this is never reached?
         debug("Sending Thread: No more data")
         return False
 
@@ -171,6 +213,52 @@ def socket_send(sock, flood=False):
 class SaxoConnectionError(Exception):
     ...
 
+regex_link = re.compile(r"(http[s]?://[^<> \"\x01]+)[,.]?")
+
+def command_path(base, cmd):
+    if ("\x00" in cmd) or (os.sep in cmd) or ("." in cmd):
+        return None
+    path = os.path.join(base, "commands", cmd)
+    if (not os.path.isfile(path)) or (not os.path.getsize(path)):
+        return None
+    return path
+
+def utf8(obj):
+    return str(obj).encode("utf-8", "replace")
+
+def utf8dict(data):
+    return {utf8(key): utf8(value) for key, value in data.items()}
+
+def process(env, cmd, path, arg):
+    path = utf8(path)
+    env = utf8dict(env)
+    octets = utf8(arg)
+
+    try: proc = subprocess.Popen([path, octets], env=env,
+        stdin=subprocess.PIPE, stdout=subprocess.PIPE)
+    except PermissionError:
+        outs = "The command file does not have executable permissions"
+    except FileNotFoundError:
+        # Might have been removed just after running this thread
+        return
+    else:
+        try: outs, errs = proc.communicate(octets + b"\n", timeout=12)
+        except subprocess.TimeoutExpired:
+            proc.kill()
+            # TODO: Use actual prefix
+            outs = "Sorry, %s took too long" % cmd
+        else:
+            outs = outs.decode("utf-8", "replace")
+            if "\n" in outs:
+                outs = outs.splitlines()[0]
+
+        code = proc.returncode or 0
+        # Otherwise: TypeError: unorderable types: NoneType() > int()
+        if (code > 0) and (not outs):
+            # TODO: Use actual prefix
+            outs = "Sorry, %s responded with an error" % cmd
+    return outs
+
 class Saxo(object):
     def __init__(self, base, opt):
         self.base = base
@@ -192,6 +280,7 @@ class Saxo(object):
         self.environment_cache["SAXO_BASE"] = base
         self.environment_cache["SAXO_COMMANDS"] = \
             os.path.join(base, "commands")
+        self.environment_cache["SAXO_VERSION"] = saxo_version
 
         self.config_cache = {}
         client_options = {
@@ -228,6 +317,7 @@ class Saxo(object):
         plugins = os.path.join(self.base, "plugins")
         sys.path[:0] = [plugins]
 
+        setups = {}
         for name in os.listdir(plugins):
             if ("_" in name) or (not name.endswith(".py")):
                 continue
@@ -252,10 +342,26 @@ class Saxo(object):
                         self.events[obj.saxo_event] = [obj]
 
                 elif hasattr(obj, "saxo_setup"):
-                    obj(self)
+                    obj.saxo_name = module.__name__ + "." + obj.__name__
+                    setups[obj.saxo_name] = obj
 
             # debug("Loaded module:", name)
 
+        debug("%s setup functions" % len(setups))
+
+        graph = {}
+        for setup in setups.values():
+            graph[setup.saxo_name] = setup.saxo_deps
+
+        database_filename = os.path.join(self.base, "database.sqlite3")
+        with sqlite.Database(database_filename) as self.db:
+            for name in common.tsort(graph):
+                debug(name)
+                if name in setups:
+                    setups[name](self)
+                else:
+                    debug("Warning: Missing dependency:", name)
+
         sys.path[:1] = []
 
     def connect(self):
@@ -276,7 +382,10 @@ class Saxo(object):
         host = self.opt["server"]["host"]
         port = int(self.opt["server"]["port"])
 
-        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        debug("Connecting to %s:%s" % (host, port))
+        # self.sock.connect((host, port))
+        self.sock = socket.create_connection((host, port))
+        # self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         if "ssl" in self.opt["server"]:
             import ssl
             debug("Warning: Using SSL, but not validating the cert!")
@@ -285,9 +394,6 @@ class Saxo(object):
                 server_side=False,
                 cert_reqs=ssl.CERT_NONE) # TODO: or CERT_REQUIRED
 
-        debug("Connecting to %s:%s" % (host, port))
-        self.sock.connect((host, port))
-
     def disconnect(self):
         # NOTE: *Do not* close the sending thread gracefully
         # The socket shutdown code below will be invoked before the queue
@@ -365,7 +471,7 @@ class Saxo(object):
                 # TODO: If the threads are still active, they should be killed
                 # Unfortunately, threads in python can't be killed
                 debug("ERROR! Unable to stop the socket threads")
-                os._exit(1)
+                exit(1)
 
         if "flood" not in self.opt["client"]:
             time.sleep(3)
@@ -389,6 +495,8 @@ class Saxo(object):
 
     def instruction_disco_receiving(self):
         self.receiving = False
+        debug("Sending disconnected event to scheduler")
+        scheduler.incoming.put(("disconnected", ()))
         if self.sending:
             outgoing.put(None)
         else:
@@ -413,7 +521,7 @@ class Saxo(object):
         debug("Found another saxo instance! %s" % pids)
         self.send("QUIT", "Another saxo instance was detected")
         self.disconnect()
-        os._exit(0)
+        exit(0)
 
     def instruction_join(self, channel):
         # NOTE: .visit can still be followed by .join
@@ -455,6 +563,14 @@ class Saxo(object):
         self.update_config("client", "channels", channels)
         self.send("PART", channel)
 
+    def instruction_periodic(self, name, period, cmd, arg, sender=None):
+        database_filename = os.path.join(self.base, "database.sqlite3")
+        with sqlite.Database(database_filename) as db:
+            p = (name, period, int(time.time()), b"scheduled",
+                 common.b64pickle((cmd, arg, sender)))
+            # TODO: This fails silently if there's a type error?
+            db["saxo_periodic"].replace(p)
+
     def instruction_ping(self):
         now = time.time()
 
@@ -482,12 +598,17 @@ class Saxo(object):
         # Could be a problem if the sender has broken
         self.send("QUIT")
         self.disconnect()
-        # sys.exit()
-        # TODO: Sometimes sys.exit doesn't work, not sure why
-        os._exit(0)
+        exit(0)
 
     def instruction_receiving(self):
         self.receiving = time.time()
+
+        scheduler.incoming.put(("connected", ()))
+        def start_scheduler():
+            scheduler.incoming.put(("start", ()))
+        start = threading.Timer(3, start_scheduler)
+        start.start()
+
         # TODO: Check that we really are connected
         # TODO: Unit test for :connected event
         incoming.put(("connected",))
@@ -542,7 +663,11 @@ class Saxo(object):
     def instruction_schedule(self, unixtime, command, args):
         command = command.encode("ascii")
         args = common.b64pickle(args)
-        scheduler.incoming.put((unixtime, command, args))
+        # TODO: Why not just add it to the database ourselves?
+        scheduler.incoming.put(("schedule.add", (unixtime, command, args)))
+
+    def instruction_scheduled(self, cmd, arg, sender=None):
+        self.scheduled_command(cmd, arg, sender=sender)
 
     def instruction_send(self, *args):
         self.send(*args)
@@ -552,49 +677,37 @@ class Saxo(object):
 
     def command(self, msg):
         cmd, arg = msg.cmd, msg.arg
-
-        if ("\x00" in cmd) or (os.sep in cmd) or ("." in cmd):
+        path = command_path(self.base, cmd)
+        if path is None:
             return
 
-        path = os.path.join(self.base, "commands", cmd)
-
-        def process(env, path, arg):
-            octets = arg.encode("utf-8", "replace")
-
-            try: proc = subprocess.Popen([path, octets], env=env,
-                stdin=subprocess.PIPE, stdout=subprocess.PIPE)
-            except PermissionError:
-                outs = "The command file does not have executable permissions"
-            except FileNotFoundError:
-                # Might have been removed just after running this thread
-                return
-            else:
-                try: outs, errs = proc.communicate(octets + b"\n", timeout=12)
-                except subprocess.TimeoutExpired:
-                    proc.kill()
-                    # TODO: Use actual prefix
-                    outs = "Sorry, %s took too long" % cmd
-                else:
-                    outs = outs.decode("utf-8", "replace")
-                    if "\n" in outs:
-                        outs = outs.splitlines()[0]
-
-                code = proc.returncode or 0
-                # Otherwise: TypeError: unorderable types: NoneType() > int()
-                if (code > 0) and (not outs):
-                    # TODO: Use actual prefix
-                    outs = "Sorry, %s responded with an error" % cmd
-
+        def command_process(env, cmd, path, arg):
+            outs = process(env, cmd, path, arg)
             if outs:
                 self.send("PRIVMSG", msg.sender, outs)
 
-        if os.path.isfile(path):
-            env = self.environment_cache.copy()
-            env["SAXO_NICK"] = msg.nick
-            env["SAXO_SENDER"] = msg.sender
-            if msg.sender in self.links:
-                env["SAXO_URL"] = self.links[msg.sender]
-            common.thread(process, env, path, arg)
+        env = self.environment_cache.copy()
+        env["SAXO_NICK"] = msg.nick
+        env["SAXO_SENDER"] = msg.sender
+        if msg.sender in self.links:
+            env["SAXO_URL"] = self.links[msg.sender]
+        if msg.authorised():
+            env["SAXO_AUTHORISED"] = "1"
+        common.thread(command_process, env, cmd, path, arg)
+
+    def scheduled_command(self, cmd, arg, sender=None):
+        path = command_path(self.base, cmd)
+        if path is None:
+            return
+
+        def command_process(env, cmd, path, arg):
+            outs = process(env, cmd, path, arg)
+            if outs and (sender is not None):
+                self.send("PRIVMSG", sender, outs)
+
+        env = self.environment_cache.copy()
+        env["SAXO_SCHEDULED"] = "1"
+        common.thread(command_process, env, cmd, path, arg)
 
     def update_config(self, section, option, value):
         self.opt[section][option] = value
@@ -612,6 +725,13 @@ class Saxo(object):
         if len(args) > 1:
             args = args[:-1] + (":" + args[-1],)
         text = re.sub(r"[\r\n]", "", " ".join(args))
+
+        if len(args) == 3:
+            if (args[0] == "PRIVMSG") and args[1].startswith("#"):
+                search = regex_link.search(args[2])
+                if search:
+                    incoming.put(("link", args[1], search.group(1)))
+
         outgoing.put(text.encode("utf-8", "replace")[:510] + b"\r\n")
 
 def serve(sockname, incoming):
@@ -666,8 +786,6 @@ def start(base):
     # TODO: Defaulting?
     # TODO: Warn if the config file is widely readable?
 
-    common.populate(saxo_path, base)
-
     sockname =  os.path.join(base, "client.sock")
     serve(sockname, incoming)
     os.chmod(sockname, 0o600)
@@ -678,7 +796,8 @@ def start(base):
             os.remove(sockname)
     atexit.register(remove_sock, sockname)
 
-    common.thread(scheduler.start, base, incoming)
+    sched = scheduler.Scheduler(incoming)
+    common.thread(sched.start, base)
 
     saxo = Saxo(base, opt)
     saxo.run()
