@@ -1,5 +1,5 @@
-# Copyright 2013-4, Sean B. Palmer
-# Source: http://inamidst.com/saxo/
+# http://inamidst.com/saxo/
+# Created by Sean B. Palmer
 
 # TODO list:
 # - Think about folding common.py into saxo.py
@@ -44,6 +44,13 @@ if path is None:
 with open(os.path.join(path, "version")) as f:
     version = f.read().rstrip("\r\n")
 
+if (os.environ.get("SAXO_VERSION") or version) != version:
+    import sys
+    message = "Error: Version mismatch: Saxo %s attempted to import Saxo %s"
+    msg = message % (version, os.environ.get("SAXO_VERSION"))
+    print(msg)
+    raise ImportError(msg)
+
 from collections import namedtuple
 
 version_info = namedtuple(
@@ -55,7 +62,7 @@ del os
 
 # Decorators:
 #
-# saxo.command(name)
+# saxo.command(*args, *, authorised=False, owner=False, private=False)
 # saxo.event(command="*")
 # saxo.pipe(function)
 # saxo.setup(function)
@@ -88,40 +95,48 @@ def client(command, *args, base=None):
     client.send(command + b" " + base64.b64encode(pickled) + b"\n")
     client.close()
 
-def command(name, owner=False):
-    def is_owner(irc):
-        import re
-        config_owner = irc.config.get("owner", "")
-        test_identity = False
-        if not "!" in config_owner:
-            test_identity = True
-            config_owner = config_owner + "!*@*"
+def command(*args, authorised=False, owner=False, private=False):
+    # TODO: arity, etc.
+    def plugin_command(name, owner=False):
+        print("Warning: Use of saxo.command is deprecated!")
 
-        mask = lambda g: "^" + re.escape(g).replace("\\*", ".*") + "$"
-        matches = re.match(mask(config_owner), irc.prefix) is not None
-        if test_identity:
-           return matches and irc.identified
-        return matches
+        def decorate(function):
+            @event("PRIVMSG")
+            def inner(irc):
+                prefix = irc.config["prefix"]
+                if irc.text.startswith(prefix):
+                    length = len(prefix)
+                    text = irc.text[length:]
 
-    def decorate(function):
-        @event("PRIVMSG")
-        def inner(irc):
-            prefix = irc.config["prefix"]
-            if irc.text.startswith(prefix):
-                length = len(prefix)
-                text = irc.text[length:]
+                    if " " in text:
+                        cmd, arg = text.split(" ", 1)
+                    else:
+                        cmd, arg = text, ""
 
-                if " " in text:
-                    cmd, arg = text.split(" ", 1)
-                else:
-                    cmd, arg = text, ""
+                    if cmd == name:
+                        irc.arg = arg
+                        if (not owner) or irc.authorised():
+                            function(irc)
+            return inner
+        return decorate
 
-                if cmd == name:
-                    irc.arg = arg
-                    if (not owner) or is_owner(irc):
-                        function(irc)
-        return inner
-    return decorate
+    if args:
+        if len(args) == 1:
+            return plugin_command(args[0], owner=owner)
+        elif (len(args) == 2) and (owner is False):
+            return plugin_command(args[0], owner=args[1])
+        else:
+            raise ValueError("Incorrect use of saxo.command")
+    elif owner is not False:
+        raise ValueError("No such keyword argument: owner")
+
+    def decorator(function):
+        if authorised and (not env("authorised")):
+            return
+        if private and env("sender").startswith("#"):
+            return
+        pipe(function)
+    return decorator
 
 # saxo_event
 # saxo_periodic
@@ -169,6 +184,7 @@ def pipe(function):
     # This gives you:
     # * concise code
     # * clean exiting
+    # * SAXO_COMMANDS check and possible setup
     # * input surrogate decoding
     # * output encoding
     # * a custom error wrapper
@@ -186,14 +202,18 @@ def pipe(function):
 
     common.exit_cleanly()
 
+    if not os.environ.get("SAXO_COMMANDS"):
+        os.environ["SAXO_COMMANDS"] = os.path.dirname(sys.argv[0])
+
     if len(sys.argv) > 1:
-        arg = sys.argv[1]
+        arg = " ".join(sys.argv[1:])
         # We have to do this because python converts arguments to surrogates
         # http://stackoverflow.com/a/7077803
         arg = os.fsencode(arg).decode("utf-8")
     elif not sys.stdin.isatty():
         arg = sys.stdin.buffer.readline()
         arg = arg.rstrip(b"\r\n")
+        arg = arg.decode("utf-8")
     else:
         arg = ""
 
@@ -233,6 +253,8 @@ def request(*args, **kargs):
 
 def setup(function):
     function.saxo_setup = True
+    if not hasattr(function, "saxo_deps"):
+        function.saxo_deps = []
     return function
 
 def script(argv):
